/*
 * Maxim Tishkov 2016
 * This file is distributed under the same license as OpenSSL
 */

#ifndef GOST_GRASSHOPPER_H
#define GOST_GRASSHOPPER_H

#if defined(__cplusplus)
extern "C" {
#endif

#include "gost_grasshopper_defines.h"
#include "gost_grasshopper_math.h"
#include "gost_grasshopper_cipher.h"
#include "gost_grasshopper_precompiled.h"
#include "gost_grasshopper_core.h"

#if defined(__cplusplus)
}
#endif

#endif
